import collections
import pathlib
import re
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from fractions import Fraction
from typing import Any

import pint

from .parser_errors import InvalidParameterError
from .tokenizer import COMMENT_CHARS, tokenize

UREG = pint.UnitRegistry()
UREG.load_definitions(str(pathlib.Path(__file__).resolve().parent.joinpath("pint_units.txt")))


def kw_converter_bool(string):
    string = string.upper()

    if string in ("0", "F", ".F.", "FALSE", ".FALSE.", "N", "NO", "OFF"):
        return False

    if string in ("1", "T", ".T.", "TRUE", ".TRUE.", "Y", "YES", "ON"):
        return True

    raise InvalidParameterError(f"invalid value given for a boolean: '{string}'")


def kw_converter_str(string):
    return string.strip("'\"")


FORTRAN_REAL = re.compile(r"(\d*\.\d+)[dD]([-+]?\d+)")


def kw_converter_float(string):
    """convert a given string to a Python float

    :param string: string with the float, can be in Fortran scientific notation or as fraction
    """
    string = FORTRAN_REAL.sub(r"\1e\2", string)

    if "/" in string:
        return float(Fraction(string))

    return float(string)


def kw_converter_keyword(string, allowed_values):
    string = string.upper()

    if string in allowed_values:
        return string

    raise InvalidParameterError(f"invalid keyword '{string}'")


KW_VALUE_CONVERTERS = {
    "logical": kw_converter_bool,
    "integer": int,
    "real": kw_converter_float,
    "word": kw_converter_str,
    "string": kw_converter_str,
}

KWDataType = collections.namedtuple("KWDataType", ["type", "n_var", "parser"])


def get_datatype(kw_node):
    dt = kw_node.find("./DATA_TYPE")
    kind = dt.get("kind")

    if kind == "keyword":
        # the keywords parser needs the list of valid keywords for verification
        valid_keywords = [e.text for e in dt.iterfind(".//NAME")]
        parser = lambda v: kw_converter_keyword(v, valid_keywords)  # noqa
    else:
        parser = KW_VALUE_CONVERTERS[kind]

    return KWDataType(kind, int(dt.find("./N_VAR").text), parser)


@dataclass
class Keyword:
    name: str
    values: Any
    repeats: bool
    node: ET.Element

    @staticmethod
    def from_string(kw_node, vstring, key_trafo=str):
        datatype = get_datatype(kw_node)

        # for a string datatype, no tokenization shall be done
        if datatype.type == "string":
            if vstring.startswith(("'", '"')):
                # if the content is actually a string, employ the tokenizer do correctly determine the end of it
                tokens = tokenize(vstring)
            else:
                tokens = [vstring.rstrip()]  # strip trailing whitespace
            # in case of no value, this will lead to an empty string, which is intentional
            # since we can't distinguish between an empty string and a lone keyword for a string datatype
            # -> inline comments for strings which are not escaped will be treated as part of the string, reproducing CP2K-behaviour
        else:
            tokens = tokenize(vstring)

        lone_keyword_value = None
        try:
            lone_keyword_value = tokenize(kw_node.find("./LONE_KEYWORD_VALUE").text)
        except AttributeError:
            pass

        if not tokens:
            if not lone_keyword_value:
                raise InvalidParameterError("keyword expects at least one value")

            tokens = lone_keyword_value

        default_unit = None
        try:
            default_unit = kw_node.find("./DEFAULT_UNIT").text
        except AttributeError:
            pass

        if default_unit and default_unit == "internal_cp2k":
            default_unit = None  # we can't do any conversion for hat

        if default_unit:
            default_unit = UREG.parse_expression(default_unit)

        current_unit = default_unit

        values = []

        for token in tokens:
            if token.startswith("["):
                if not default_unit:
                    raise InvalidParameterError(
                        "unit specified for value in keyword, but no default unit available or default unit is 'internal_cp2k'"
                    )
                current_unit = UREG.parse_expression(token.strip("[]"), case_sensitive=False)
                continue

            if token.startswith(COMMENT_CHARS):
                assert token == tokens[-1], "found inline comment which is not the last token"
                continue  # ignore inline comments

            value = datatype.parser(token)

            if datatype.type == "keyword":
                # keywords are also matched case insensitive, apply the same rules as for the keys
                value = key_trafo(value)

            if current_unit != default_unit:
                # interpret the given value in the specified unit, convert it and get the raw value
                value = (value * current_unit).to(default_unit).magnitude

            values += [value]

        if not values:
            raise InvalidParameterError("keyword expects at least one value, only a unit spec was given")

        if (datatype.n_var > 0) and (datatype.n_var != len(values)):
            raise InvalidParameterError(f"keyword expects exactly {datatype.n_var} values, {len(values)} were given")

        # simplify the value if only one is given/requested
        if len(values) == 1:
            values = values[0]
        else:
            values = tuple(values)

        key_name = kw_node.find("./NAME[@type='default']").text
        if key_name == "DEFAULT_KEYWORD":
            key_name = "*"

        return Keyword(key_name, values, True if kw_node.get("repeats") == "yes" else False, kw_node)



import contextlib
import pathlib
import sys

import click


@contextlib.contextmanager
def smart_open(filename=None, mode="r"):
    """A context manager to automatically read from stdin, based on https://stackoverflow.com/a/17603000/1400465"""

    assert mode in ("r", "w", "x"), "Only r and w/x modes are supported"

    filebased = filename and str(filename) != "-"

    if filebased and mode == "r":
        fhandle = open(filename, "r")
    elif filebased and mode in ("x", "w"):
        fhandle = open(filename, mode)
    elif mode == "r":
        fhandle = sys.stdin
    else:
        fhandle = sys.stdout

    try:
        yield fhandle
    finally:
        if filebased:
            fhandle.close()


def click_validate_kv(ctx, param, args):
    if isinstance(args, dict):
        return args

    parsed = {}

    for arg in args:
        try:
            parsed.update((arg.split("=", maxsplit=1),))
        except ValueError:
            raise click.BadParameter(f"format must be 'key=value', got '{arg}'") from None

    return parsed


def fhandle_argument(func):
    return click.argument("fhandle", metavar="[<file>]", type=click.File(), default="-")(func)


def yaml_option(func):
    return click.option("--yaml", "-y", is_flag=True, help="generate YAML output instead of the default JSON")(func)


def canonical_option(func):
    return click.option("-c", "--canonical", is_flag=True, help="use the canonical output format instead of the simplified one")(
        func
    )


def base_dir_option(func):
    return click.option(
        "-b",
        "--base-dir",
        type=click.Path(exists=True, file_okay=False, path_type=pathlib.Path),
        default=".",
        help="search path used for relative @include's",
        show_default=True,
    )(func)


def var_values_option(func):
    return click.option(
        "var_values",
        "-E",
        "--set",
        metavar="key=value",
        multiple=True,
        type=click.UNPROCESSED,
        callback=click_validate_kv,
        help="preset the value for a CP2K preprocessor variable",
    )(func)



import functools
import json
import sys
from enum import Enum
from typing import Mapping, MutableSequence

import click

from cp2k_input_tools.parser import (
    CP2KInputParser,
    CP2KInputParserAiiDA,
    CP2KInputParserSimplified,
)

from . import base_dir_option, canonical_option, fhandle_argument, var_values_option


def _key_trafo(string):
    if len(string) <= 3:
        return string.upper()
    return string.lower()


class Trafos(Enum):
    # see https://stackoverflow.com/a/40486992 need to wrap functions in function objects
    auto = functools.partial(_key_trafo)
    lower = functools.partial(str.lower)
    upper = functools.partial(str.upper)


@click.command()
@fhandle_argument
@click.option(
    "oformat", "-f", "--format", type=click.Choice(("json", "yaml", "aiida-cp2k-calc")), default="json", help="output format"
)
@canonical_option
@base_dir_option
@click.option(
    "-t",
    "--trafo",
    type=click.Choice([t.name for t in Trafos]),
    callback=lambda c, p, v: getattr(Trafos, v) if v else None,
    default="auto",  # see https://github.com/pallets/click/issues/605#issuecomment-847361079
    help="transformation applied to key and section names",
)
@var_values_option
def fromcp2k(fhandle, oformat, canonical, base_dir, trafo, var_values):
    """Convert CP2K input to JSON (default), YAML or an aiida-cp2k run script template"""

    if oformat == "aiida-cp2k-calc":
        if canonical:
            print("The --canonical argument is ignored when generating an aiida-cp2k run script template", file=sys.stderr)
        if trafo != Trafos.auto:
            print(
                "Any key transformation function other than 'auto' is ignored when generating an aiida-cp2k run script template",
                file=sys.stderr,
            )
        cp2k_parser = CP2KInputParserAiiDA(base_dir=base_dir)
    elif canonical:
        cp2k_parser = CP2KInputParser(base_dir=base_dir, key_trafo=trafo.value)
    else:
        cp2k_parser = CP2KInputParserSimplified(base_dir=base_dir, key_trafo=trafo.value)

    tree = cp2k_parser.parse(fhandle, dict(var_values))

    if oformat == "json":
        print(json.dumps(tree, indent=2))

    elif oformat == "yaml":
        from ruamel.yaml import YAML

        yaml = YAML()
        yaml.dump(tree, sys.stdout)

    elif oformat == "aiida-cp2k-calc":
        from jinja2 import Environment, PackageLoader

        env = Environment(loader=PackageLoader("cp2k_input_tools", "templates"))
        env.globals.update({"isinstance": isinstance, "Mapping": Mapping, "MutableSequence": MutableSequence})
        env.filters["quoted"] = lambda item: f'"{item}"' if isinstance(item, str) else item
        template = env.get_template("aiida_cp2k_calc.py.j2")
        print(template.render(tree=tree))



from dataclasses import dataclass, field
from typing import Any, List, Optional, Tuple

import transitions

COMMENT_CHARS = ("!", "#")


class TokenizerError(Exception):
    pass


class UnterminatedStringError(TokenizerError):
    pass


class InvalidTokenCharError(TokenizerError):
    pass


@dataclass
class Context:
    colnr: Optional[int] = None
    colnrs: List[int] = field(default_factory=list)
    ref_colnr: Optional[int] = None
    line: Optional[str] = None
    ref_line: Optional[str] = None
    filename: Optional[str] = None
    section: Any = None


@dataclass
class Token:
    string: str
    ctx: Context


class CP2KInputTokenizer(transitions.Machine):
    def begin_basic_token(self, _, colnr):
        self._current_token_start = colnr

    def end_basic_token(self, _, colnr: int):
        # the end idx follows the python-default of specifying ranges,
        # since this is triggered on the character after, using idx is correct
        self._tokens += [(self._current_token_start, colnr)]

    def begin_string_token(self, content, colnr):
        self._current_token_start = colnr
        self._tracking_quote_char = content[colnr]

    def end_string_token(self, content, colnr):
        # this is trigger ON the encounter of the string token, while the
        # end of the basic token is determined by the character that follows
        self._tokens += [(self._current_token_start, colnr + 1)]

    def unterminated_string(self, _, colnr):
        raise UnterminatedStringError("unterminated string detected", Context(colnr=colnr, ref_colnr=self._current_token_start))

    def invalid_token_char(self, content, colnr):
        raise InvalidTokenCharError("invalid keyword character found", Context(colnr=colnr, ref_colnr=self._current_token_start))

    def is_not_escaped(self, content, colnr):
        if colnr > 0:
            # possible to do: account for multiple escapes
            return content[colnr - 1] != "\\"

        return True

    @property
    def tokens(self):
        return self._tokens

    def is_matching_quote(self, content, colnr):
        return self._tracking_quote_char == content[colnr]

    def __init__(self):
        super().__init__(
            self,
            initial="lookout",
            states=[
                transitions.State(name="lookout"),
                transitions.State(name="basic_token", on_enter=["begin_basic_token"], on_exit=["end_basic_token"]),
                transitions.State(name="string_token", on_enter=["begin_string_token"], on_exit=["end_string_token"]),
                transitions.State(name="comment", on_enter=["begin_basic_token"], on_exit=["end_basic_token"]),
            ],
            transitions=[
                # start parsing a token:
                {"trigger": "token_char", "source": "lookout", "dest": "basic_token"},
                # ... unless we're already parsing a token or inside a string or comment
                {"trigger": "token_char", "source": ["basic_token", "string_token", "comment"], "dest": None},
                # '/" initiate strings
                {"trigger": "quote_char", "source": "lookout", "dest": "string_token"},
                {
                    "trigger": "quote_char",
                    "source": "string_token",
                    "dest": "lookout",
                    "conditions": ["is_not_escaped", "is_matching_quote"],
                },
                # a '!' or '#' initiates a comment (and terminates a token if necessary)
                {"trigger": "comment_char", "source": ["lookout", "basic_token"], "dest": "comment"},
                # ... unless inside a single or double quoted string, where it is consumed:
                {"trigger": "comment_char", "source": "string_token", "dest": None},
                # whitespace terminates a basic token
                {"trigger": "ws_char", "source": "basic_token", "dest": "lookout"},
                # ... and is consumed in all other cases
                {"trigger": "ws_char", "source": ["lookout", "string_token", "comment"], "dest": None},
                # single/double quotes are not allowed in a basic token:
                {"trigger": "quote_char", "source": "basic_token", "before": "invalid_token_char", "dest": None},
                {"trigger": "nl_char", "source": ["basic_token", "comment"], "dest": "lookout"},
                {"trigger": "nl_char", "source": "lookout", "dest": None},
                {"trigger": "nl_char", "source": "string_token", "before": "unterminated_string", "dest": None},
            ],
        )

        self._tracking_quote_char = None
        self._current_token_start = 0
        self._tokens = []


def tokenize(string: str) -> Tuple[str, ...]:
    tokenizer = CP2KInputTokenizer()

    char_map = {" ": tokenizer.ws_char, "\t": tokenizer.ws_char, "'": tokenizer.quote_char, '"': tokenizer.quote_char}

    for cchar in COMMENT_CHARS:
        char_map[cchar] = tokenizer.comment_char

    for colnr, char in enumerate(string):
        char_map.get(char, tokenizer.token_char)(string, colnr)

    tokenizer.nl_char(string, len(string))

    return tuple(string[s:e] for s, e in tokenizer.tokens)



import re
from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from typing import NamedTuple, Sequence

from .lineiterator import MultiFileLineIterator
from .parser_errors import PreprocessorError
from .tokenizer import COMMENT_CHARS, Context, TokenizerError, tokenize


class _Variable(NamedTuple):
    value: str
    ctx: defaultdict


class _ConditionalBlock(NamedTuple):
    condition: str
    ctx: defaultdict


_VALID_VAR_NAME_MATCH = re.compile(r"^[a-z_]\w*$", flags=re.IGNORECASE | re.ASCII)
_CONDITIONAL_MATCH = re.compile(r"\s*@(?P<stmt>IF|ENDIF)\s*(?P<cond>.*)", flags=re.IGNORECASE)
_SET_MATCH = re.compile(r"\s*@SET\s+(?P<var>\S+)\s+(?P<value>.+)", flags=re.IGNORECASE)
_INCLUDE_MATCH = re.compile(r"\s*(?P<complete>@(?P<type>INCLUDE|XCTYPE)\b\s*(?P<file>.*))", flags=re.IGNORECASE)


class CP2KPreprocessor(Iterator):
    def __init__(self, fhandle, base_dir, initial_variable_values=None):
        self._varstack = {}
        self._lineiter = MultiFileLineIterator()
        self._conditional_block = None
        self._current_line_entry = None

        if isinstance(base_dir, (str, Path)):
            self._inc_dirs = [Path(base_dir)]
        elif isinstance(base_dir, Sequence):
            self._inc_dirs = [Path(b) for b in base_dir]
        else:
            raise TypeError("invalid type passed for base_dir")

        if initial_variable_values:
            self._varstack.update({k.upper(): _Variable(v, None) for k, v in initial_variable_values.items()})

        self._lineiter.add_file(fhandle, managed=False)

    def _resolve_variables(self, line):
        var_start = 0
        var_end = 0

        ctx = Context(line=line)

        # the following algorithm is from CP2Ks cp_parser_inpp_methods.F to reproduce its behavior :(

        # first replace all "${...}"  with no nesting, meaning that ${foo${bar}} means foo$bar is the key
        while True:
            var_start = line.find("${")
            if var_start < 0:
                break

            var_end = line.find("}", var_start + 2)
            if var_end < 0:
                ctx.colnr = len(line)
                ctx.ref_colnr = var_start
                raise PreprocessorError("unterminated variable", ctx)

            ctx.colnr = var_start
            ctx.ref_colnr = var_end

            key = line[var_start + 2 : var_end]  # without ${ and }
            value = None

            try:
                # see whether we got a default value and unpack
                key, value = key.split("-", maxsplit=1)
            except ValueError:
                pass

            if not _VALID_VAR_NAME_MATCH.match(key):
                raise PreprocessorError(f"invalid variable name '{key}'", ctx) from None

            try:
                value = self._varstack[key.upper()].value
            except KeyError:
                if value is None:
                    raise PreprocessorError(f"undefined variable '{key}' (and no default given)", ctx) from None

            line = f"{line[:var_start]}{value}{line[var_end+1:]}"

        var_start = 0
        var_end = 0

        while True:
            var_start = line.find("$")
            if var_start < 0:
                break

            var_end = line.find(" ", var_start + 1)
            if var_end < 0:
                # -1 would be the last entry, but in a range it is without the specified entry
                var_end = len(line.rstrip())

            ctx.colnr = var_start
            ctx.ref_colnr = var_end - 1

            key = line[var_start + 1 : var_end]

            if not _VALID_VAR_NAME_MATCH.match(key):
                raise PreprocessorError(f"invalid variable name '{key}'", ctx) from None

            try:
                value = self._varstack[key.upper()].value
            except KeyError:
                raise PreprocessorError(f"undefined variable '{key}'", ctx) from None

            line = f"{line[:var_start]}{value}{line[var_end:]}"

        return line

    def _parse_preprocessor_instruction(self, line):
        conditional_match = _CONDITIONAL_MATCH.match(line)

        ctx = Context(line=line)

        if conditional_match:
            stmt = conditional_match.group("stmt")
            condition = conditional_match.group("cond").strip()

            if stmt.upper() == "ENDIF":
                if self._conditional_block is None:
                    raise PreprocessorError("found @ENDIF without a previous @IF", ctx)

                # check for garbage which is not a comment, note: we're stricter than CP2K here
                if condition and not condition.startswith(COMMENT_CHARS):
                    ctx.colnr = conditional_match.start("cond")
                    ctx.ref_colnr = conditional_match.end("cond")
                    raise PreprocessorError("garbage found after @ENDIF", ctx)

                self._conditional_block = None
            else:
                if self._conditional_block is not None:
                    ctx.ref_line = self._conditional_block.ctx.line
                    raise PreprocessorError("nested @IF are not allowed", ctx)

                # resolve any variables inside the condition
                try:
                    condition = self._resolve_variables(condition)
                except PreprocessorError as exc:
                    exc.args[1].colnr += conditional_match.start("cond")
                    exc.args[1].ref_colnr += conditional_match.start("cond")
                    raise

                # prefix-whitespace are consumed in the regex, suffix with the strip() above
                if not condition or condition == "0":
                    self._conditional_block = _ConditionalBlock(False, ctx)
                elif "==" in condition:
                    lhs, rhs = [s.strip() for s in condition.split("==", maxsplit=1)]
                    self._conditional_block = _ConditionalBlock(lhs == rhs, ctx)
                elif "/=" in condition:
                    lhs, rhs = [s.strip() for s in condition.split("/=", maxsplit=1)]
                    self._conditional_block = _ConditionalBlock(lhs != rhs, ctx)
                else:
                    self._conditional_block = _ConditionalBlock(True, ctx)

            return

        if self._conditional_block and not self._conditional_block.condition:
            return

        set_match = _SET_MATCH.match(line)
        if set_match:
            # resolve other variables in the definition first
            key = set_match.group("var")
            value = self._resolve_variables(set_match.group("value"))

            if not _VALID_VAR_NAME_MATCH.match(key):
                raise PreprocessorError(f"invalid variable name '{key}'", ctx) from None

            self._varstack[key.upper()] = _Variable(value, ctx)
            return

        include_match = _INCLUDE_MATCH.match(line)
        if include_match:
            inctype = include_match["type"]

            # resolve variables first
            try:
                filename = self._resolve_variables(include_match.group("file"))
            except PreprocessorError as exc:
                exc.args[1].colnr += include_match.start("file")  # shift colnr
                exc.args[1].ref_colnr += include_match.start("file")
                raise

            if filename.startswith(("'", '"')):
                try:
                    tokens = tokenize(filename)  # use the tokenizer to detect unterminated quotes
                except TokenizerError as exc:
                    exc.args[1].colnr += include_match.start("file")  # shift colnr
                    exc.args[1].ref_colnr += include_match.start("file")
                    raise

                if len(tokens) != 1:
                    raise PreprocessorError(
                        "@INCLUDE requires exactly one argument",
                        Context(colnr=include_match.start("complete"), ref_colnr=include_match.end("complete")),
                    )

                filename = tokens[0].strip("'\"")

            if not filename:
                raise PreprocessorError(
                    f"@{inctype} requires exactly one argument",
                    Context(colnr=include_match.start("complete"), ref_colnr=include_match.end("complete")),
                )

            filename = filename.strip("'\"")

            if inctype.upper() == "XCTYPE":
                filename = f"xc_section/{filename}.sec"

            for inc_dir in self._inc_dirs:
                try:
                    # if the filename is an absolute path, joinpath uses that one and will ignore the dir
                    fhandle = inc_dir.joinpath(filename).open("r")

                    # the _lineiter takes over the handle and closes it at EOF
                    self._lineiter.add_file(fhandle)

                    break
                except OSError:
                    continue
            else:
                raise PreprocessorError(f"specified INCLUDE/XCTYPE {filename} could not be opened", ctx)

            return

        raise PreprocessorError("unknown preprocessor directive found", ctx)

    def __next__(self):
        for line in self._lineiter:
            try:
                # ignore empty lines and comments:
                if not line or line.startswith(COMMENT_CHARS):
                    continue

                if line.startswith("@"):
                    self._parse_preprocessor_instruction(line)
                    continue

                # ignore everything in a disable @IF/@ENDIF block
                if self._conditional_block and not self._conditional_block.condition:
                    continue

                return self._resolve_variables(line)

            except (PreprocessorError, TokenizerError) as exc:
                exc.args[1].filename = self._lineiter.fname
                exc.args[1].linenr = self._lineiter.line_range[1]
                exc.args[1].colnrs = self._lineiter.colnrs
                exc.args[1].line = line
                raise

        if self._conditional_block is not None:
            raise PreprocessorError(
                "conditional block not closed at end of file", Context(ref_line=self._conditional_block.ctx.line)
            )

        raise StopIteration

    @property
    def line_range(self):
        """Original line numbers (start and end) of the last (possibly combined) line"""
        return self._lineiter.line_range

    @property
    def colnrs(self):
        """Original column numbers where non-whitespace content started for most recent emitted line"""
        return self._lineiter.colnrs

    @property
    def starts(self):
        """Index in the most recent emitted line where content from a new line in file starts"""
        return self._lineiter.starts

    @property
    def fname(self):
        return self._lineiter.fname



import json

import click

from cp2k_input_tools.generator import CP2KInputGenerator

from . import fhandle_argument, yaml_option


@click.command()
@fhandle_argument
@yaml_option
def tocp2k(fhandle, yaml):
    """
    Generate a CP2K input file based on JSON (or YAML) file
    """

    if yaml:
        from ruamel.yaml import YAML

        yaml = YAML()

        tree = yaml.load(fhandle)
    else:

        tree = json.load(fhandle)

    cp2k_generator = CP2KInputGenerator()

    for line in cp2k_generator.line_iter(tree):
        print(line)



from collections.abc import Iterator
from typing import List, NamedTuple, TextIO, Tuple


class LineContinuationError(Exception):
    pass


class ContinuationLineIterator(Iterator):
    def __init__(self, fhandle):
        self._fiter = enumerate(fhandle, start=1)
        self._linenr = 0
        self._colnrs = []  # number of stripped whitespace (from current line, continued line, 2nd continued line, ...)
        self._starts = [0]  # list of newline starts (in case of line continuations)

    @property
    def line_range(self) -> Tuple[int, int]:
        """Original line numbers (start and end) of the last (possibly combined) line"""
        return (self._linenr - len(self._starts) - 1, self._linenr)

    @property
    def colnrs(self) -> List[int]:
        """Original column numbers where non-whitespace content started for most recent emitted line"""
        return self._colnrs

    @property
    def starts(self) -> List[int]:
        """Index in the most recent emitted line where content from a new line in file starts"""
        return self._starts

    def __next__(self) -> str:
        """yields LineEntries"""

        self._starts = [0]
        self._colnrs = []
        line = ""

        for self._linenr, raw_line in self._fiter:
            lstripped = raw_line.lstrip()  # CP2K consequently strips all left whitespace
            self._colnrs += [len(raw_line) - len(lstripped)]  # remember where the original colnr started

            # Python has universal line endings (always only \n)
            line += lstripped.rstrip("\n")

            if line.endswith("\\"):
                line = line[:-1]
                self._starts += [len(line)]
                continue

            return line

        if self._colnrs:
            raise LineContinuationError("stray line continuation at end of file")

        raise StopIteration


class _FileIterPair(NamedTuple):
    fhandle: TextIO
    iter: ContinuationLineIterator
    managed: bool = True  # whether the iterator should close it at EOF


class MultiFileLineIterator(Iterator):
    def __init__(self):
        self._fileiterpairs = []

    def __del__(self):
        while self._fileiterpairs:
            self._fileiterpairs.pop().fhandle.close()

    def add_file(self, fhandle, managed=True):
        self._fileiterpairs += [_FileIterPair(fhandle, ContinuationLineIterator(fhandle), managed)]

    def __next__(self) -> str:
        while self._fileiterpairs:
            try:
                # yield the next (possibly continued) line from the file on top of the stack
                return next(self._fileiterpairs[-1].iter)
            except StopIteration:
                # end of current file, remove the file from the stack and close it
                fhandle, _, managed = self._fileiterpairs.pop()
                if managed:
                    fhandle.close()

        raise StopIteration

    @property
    def line_range(self) -> Tuple[int, int]:
        """Original line numbers (start and end) of the last (possibly combined) line"""
        return self._fileiterpairs[-1].iter.line_range

    @property
    def colnrs(self) -> List[int]:
        """Original column numbers where non-whitespace content started for most recent emitted line"""
        return self._fileiterpairs[-1].iter.colnrs

    @property
    def starts(self) -> List[int]:
        """Index in the most recent emitted line where content from a new line in file starts"""
        return self._fileiterpairs[-1].iter.starts

    @property
    def fname(self) -> str:
        return getattr(self._fileiterpairs[-1].fhandle, "name", "<BUFFER>")



import collections
import itertools
import re
import xml.etree.ElementTree as ET

from . import DEFAULT_CP2K_INPUT_XML


class GeneratorError(Exception):
    pass


class SectionNotFoundError(Exception):
    pass


class KeywordNotFoundError(Exception):
    pass


class SectionParametersNotFoundError(Exception):
    pass


class InvalidSectionDataError(Exception):
    pass


class InvalidKeywordDataError(Exception):
    pass


class InvalidBooleanDataError(Exception):
    pass


class SimplifiedSectionAmbiguityError(Exception):
    pass


TreeNode = collections.namedtuple("TreeNode", ["name", "dictref", "xmlnode", "indent"])


class CP2KInputGenerator:
    def __init__(self, xmlspec=DEFAULT_CP2K_INPUT_XML, indent_shift=3):
        """Initialize the generator

        :param xmlspec: path to a `cp2k_input.xml` file (default: use the bundled)
        :param indent_shift: Number of spaces to use for indentation of the
                             keywords/subsections of a section
        """
        self._parse_tree = ET.parse(xmlspec)
        self._shift = indent_shift

    def _get_section(self, name, node):
        for section in node.iterfind("./SECTION"):
            if name.upper() in [e.text for e in section.iterfind("./NAME")]:
                break
        else:  # if we come to the end without a break:
            raise SectionNotFoundError()

        return section

    def _get_keyword(self, name, section_node):

        for kw in section_node.iterfind("./KEYWORD"):
            if name.upper() in [e.text for e in kw.iterfind("./NAME")]:
                break
        else:  # if we come to the end without a break:
            raise KeywordNotFoundError(f"keyword {name} not allowed")

        return kw

    def _get_section_parameter(self, section_node):
        # there is only one
        sparam = section_node.find("./SECTION_PARAMETERS")

        if sparam:
            return sparam

        raise SectionParametersNotFoundError("section does not take any parameters")

    def _get_default_keyword(self, section_node):
        # there is only one
        dkw = section_node.find("./DEFAULT_KEYWORD")

        if dkw:
            return dkw

        raise GeneratorError("invalid default keyword")

    def _render_keyword(self, value, kw_node):
        repeats = True if kw_node.get("repeats") == "yes" else False
        dt = kw_node.find("./DATA_TYPE")
        kind = dt.get("kind")
        n_var = int(dt.find("./N_VAR").text)

        def word_renderer(string):
            # words with whitespaces need to be quoted
            if re.search(r"\s", string):
                return f'"{string}"'

            return string

        def bool_renderer(val):
            """the given value may or may not be a boolean, make sure we also accept other datatypes"""

            if str(val).upper() in ("1", "T", ".T.", "TRUE", ".TRUE.", "Y", "YES", "ON"):
                return ".TRUE."

            if str(val).upper() in ("0", "F", ".F.", "FALSE", ".FALSE.", "N", "NO", "OFF"):
                return ".FALSE."

            raise InvalidBooleanDataError(f"the given value '{val}' could not be recognized as a CP2K boolean value")

        TYPE_RENDERERS = {
            "integer": lambda v: str(v),
            "keyword": lambda v: str(v).upper(),
            "logical": bool_renderer,
            "real": lambda v: str(v),
            "string": lambda v: v,
            "word": word_renderer,
        }

        if isinstance(value, (list, tuple)):
            if not repeats and n_var == 1:
                raise InvalidKeywordDataError("got multiple values for non-repeating single-valued keyword")

            nested_list = any(isinstance(v, (list, tuple)) for v in value)

            if repeats and n_var == 1 and nested_list:
                raise InvalidKeywordDataError("got multiple values for repeating single-valued keyword")

            # if any of the values is a list itself we know the outer list to be for the repetion and the inner for the values
            # and we only have to ensure that those entries where the inner list is not a list but a single word are also a list
            if nested_list:
                value = [v if isinstance(v, (list, tuple)) else [v] for v in value]

            elif n_var == 1:
                value = [[v] for v in value]

            else:
                # for the case where none of the values are actually list entries, we give priority to the values, meaning that
                #   "BASIS_SET": ["ORB", "pob-TZVP"]
                # becomes
                #   BASIS_SET ORB pob_TZVP
                # So if one wants to have
                #   BASIS_SET pob-TZVP
                #   BASIS_SET some-added-basis
                # s/he must explicitly provide it as:
                #   "BASIS_SET": [["pob-TZVP"], ["some-added-basis"]]
                # to break ambiguity
                value = [value]

        else:
            # replace by list of lists, because that's the most general thing
            value = [[value]]

        for entry in value:
            yield " ".join(TYPE_RENDERERS[kind](v) for v in entry)

    def _parse_section(self, name, node, path, content, indent):
        repeats = True if node.get("repeats") == "yes" else False

        # if a section contains a list, the only option is that it is a
        if isinstance(content, list):
            if not repeats:
                raise InvalidSectionDataError("multiple entries given for non-repeating section")

            if not all(isinstance(v, dict) for v in content):
                raise InvalidSectionDataError("non-section value given in repeating section")

            return [TreeNode(path + [name], v, node, indent + self._shift) for v in content]

        if not isinstance(content, dict):
            # if the value does not match the one expected for a section
            raise InvalidSectionDataError()

        section_params = None
        try:
            section_params = self._get_section_parameter(node)
        except SectionParametersNotFoundError:
            pass

        # if we have a section taking section parameters, we also accept the format:
        #   kind:
        #     H:
        #       basis_set: ...
        # instead of:
        #   kind:
        #     - _: H
        #       basis_set: ...
        # If all of the values in this dict are dicts themselves we therefore have to look closer,
        # since this could either be a simplified section (like the one above) or a regular one, like:
        #   BS:
        #     alpha:
        #       ...
        #     beta:
        #       ...
        if section_params and all(isinstance(v, dict) for v in content.values()):
            # 1. figure out all the valid keys in simplified and canonical mode
            valid_keys = [kw.text for kw in itertools.chain(node.iterfind("./KEYWORD/NAME"), node.iterfind("./SECTION/NAME"))]
            valid_keys += [f"+{kw.text}" for kw in node.iterfind("./SECTION/NAME")]

            # 2. if all of the specified keys match either section or a keyword, this is just a regular section
            if all(str(k).upper() in valid_keys for k in content.keys()):
                return [TreeNode(path + [name], content, node, indent + self._shift)]

            # 3. if not any of the keys is a valid key, we can safely assume that the user specified simplified expanded section
            if not any(str(k).upper() in valid_keys for k in content.keys()):
                # return a list of sections with this param merged into the section as normal section parameter
                return [
                    TreeNode(path + [name], dict(_=param, **section), node, indent + self._shift)
                    for param, section in content.items()
                ]

            # otherwise the user has either specified a mix (not supported) or has a typo
            raise SimplifiedSectionAmbiguityError(
                f"the section '{name}' contains elements which could either be"
                f" keywords/subsections or section parameters within the simplified"
                f" format scheme. Please use the canonical format for this section."
            )

        return [TreeNode(path + [name], content, node, indent + self._shift)]

    def line_iter(self, tree):
        """
        A generator for CP2K input file lines

        :param tree: a canonical or simplified input option tree,
                     as generated by `CP2KInputParser` or `CP2KInputParserSimplified`
        """

        treerefs = [TreeNode([""], tree, self._parse_tree.getroot(), -self._shift)]

        while treerefs:
            path, content, node, indent = treerefs.pop()
            sname = path[-1]

            if sname:
                if "_" in content:  # default parameter is a keyword "_"
                    sparam = self._get_section_parameter(node)
                    # the default parameter can only be mentioned once, so _render_keyword will only yield once
                    yield f"{str():>{indent}}&{sname} {next(self._render_keyword(content['_'], sparam))}"
                else:
                    yield f"{str():>{indent}}&{sname}"

            for key, value in content.items():
                section = None

                if key.startswith(("&", "+")):  # a section can be explicitly tagged by a starting + or &
                    section_name = key[1:].upper()
                else:
                    section_name = key.upper()

                try:
                    # if key being parsed points to a section (resp. multiple repeated sections, add them to the stack)
                    section = self._get_section(section_name, node)
                    treerefs += self._parse_section(section_name, section, path, value, indent)
                    continue
                except SectionNotFoundError:
                    # if no section name was found then go on parsing it as a key
                    pass
                except InvalidSectionDataError as exc:
                    # if the value for this key didn't match the one expected for a section
                    # check whether there is also a key with the same name and if there is, continue
                    try:
                        self._get_keyword(section_name, node)
                    except KeywordNotFoundError:
                        raise exc from None

                if key == "_":  # section parameters are handled above
                    continue

                if key == "*":
                    keyword = self._get_default_keyword(node)

                    for valuestr in self._render_keyword(value, keyword):
                        yield f"{str():>{indent + self._shift}}{valuestr}"

                    continue

                key_name = key.upper()

                keyword = self._get_keyword(key_name, node)
                for valuestr in self._render_keyword(value, keyword):
                    yield f"{str():>{indent + self._shift}}{key_name} {valuestr}"

            if treerefs:
                next_indent = treerefs[-1].indent
            else:
                next_indent = 0

            for eindent in range(indent, next_indent - 1, -self._shift):
                parent_sname = path.pop()
                yield f"{str():>{eindent}}&END {parent_sname}"



# Copyright (c), Tiziano Müller
# SPDX-License-Identifier: MIT

import itertools
import re
import warnings
from dataclasses import dataclass
from typing import (
    IO,
    Any,
    Callable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Protocol

SYM2NUM = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Fl": 114,
    "Lv": 116,
}

EOF_MARKER_LINE = "Eof marker"
EMPTY_LINE_MATCH = re.compile(r"^(\s*|\s*#.*)$")
BLOCK_MATCH = re.compile(r"^\s*(?P<element>[a-zA-Z]{1,3})\s+(?P<family>\S+).*")


def chained_exception(cls: type, msg: str, prev: Exception):
    """Create a chained exception"""
    exc = cls(msg)
    exc.__cause__ = prev
    return exc


class MulitpleValueErrorsException(ValueError):
    """An exception which contains multiple value error exceptions"""


_T = TypeVar("_T")


class SupportsFromLines(Protocol):
    @classmethod
    def from_lines(cls: Type[_T], lines: Sequence[str]) -> _T:
        pass


@dataclass
class DatafileIterMixin:
    @classmethod
    def datafile_iter(
        cls: SupportsFromLines, content: Union[IO, str, Sequence[str]], keep_going=True, emit_comments=False
    ) -> Iterator[Union[SupportsFromLines, str]]:
        """
        Generates a sequence of data instances, one for each data entry found in the given file

        :param fhandle: Open file handle (in text mode) to a data file
        :param keep_going: Whether to ignore invalid entries and keep going
        :param emit_comments: Whether to also emit line breaks and comments when found
        """

        # find the beginning of a new data entry, then
        # continue until the next one, the iterator chain and Eof marker guarantee
        # that we find a last (empty) one which will not be parsed.

        line_buffer: List[str] = []
        errors: List[Exception] = []
        post_comments: List[str] = []

        if isinstance(content, str):
            content = content.splitlines()

        for line in itertools.chain(content, [EOF_MARKER_LINE]):
            if EMPTY_LINE_MATCH.match(line):
                if line_buffer:
                    post_comments.append(line.strip())
                elif emit_comments:
                    yield (line.strip())
                continue  # ignore empty and comment lines

            match = BLOCK_MATCH.match(line)

            if match and line_buffer:
                try:
                    yield cls.from_lines(line_buffer)
                except ValueError as exc:
                    newexc = chained_exception(ValueError, f"failed to parse block for: {line_buffer[0].strip()}", exc)
                    if not keep_going:
                        raise newexc
                    errors.append(newexc)

                if post_comments:
                    if emit_comments:
                        yield from post_comments
                    post_comments = []

                line_buffer = []

            line_buffer.append(line.strip())

        assert len(line_buffer) == 1 and line_buffer[0] == EOF_MARKER_LINE

        if post_comments and emit_comments:
            yield from post_comments

        if len(errors) == 1:
            raise errors[0]

        if errors:
            raise MulitpleValueErrorsException(errors)


class FromDictMixin:
    @classmethod
    def from_dict(cls: Type[_T], data: Mapping[str, Any], type_hooks: Optional[Mapping[Type, Callable[[Any], Any]]] = None) -> _T:
        """Create a data instance from a nested dictionary"""
        warnings.warn("This helper function will be removed, use '.parse_obj' instead.", PendingDeprecationWarning)
        if type_hooks:
            warnings.warn("The 'type_hooks' attribute has been removed and is being ignored.", DeprecationWarning)

        return cls.parse_obj(data)  # type: ignore


def dformat(val, ndigits, slen):
    """
    Right-pads a decimal with spaces such that there are max_exp number of characters after the dot
    and the complete string is max_len characters in width.
    """

    digits = ndigits + val.as_tuple().exponent if val.as_tuple().exponent < 0 else ndigits + 1
    return f"{format(val, 'f') + ' '*(digits):>{slen}}"



import itertools
import pathlib
import re
import sys
from copy import deepcopy

import click

from cp2k_input_tools.generator import CP2KInputGenerator
from cp2k_input_tools.parser import CP2KInputParser, CP2KInputParserSimplified

from .. import __version__
from . import base_dir_option, canonical_option, fhandle_argument, var_values_option


@click.command()
@fhandle_argument
@click.argument("expressions", metavar="<expression>", type=str, nargs=-1)
@base_dir_option
@canonical_option
@var_values_option
def cp2kgen(fhandle, expressions, base_dir, canonical, var_values):
    """
    Generates variations of the given CP2K input file

    Examples for generator expressions:

    "force_eval/dft/mgrid/cutoff=[300,400,500,800]", "force_eval/subsys/cell/a/0=10.0"
    """

    if canonical:
        cp2k_parser = CP2KInputParser(base_dir=base_dir, key_trafo=str.lower)
    else:
        cp2k_parser = CP2KInputParserSimplified(base_dir=base_dir, key_trafo=str.lower)

    tree = cp2k_parser.parse(fhandle, dict(var_values))

    # list of substitutions/transformations to apply
    substitutions = []

    for expression in expressions:
        try:
            kpath, value = expression.split("=", maxsplit=1)
        except ValueError:
            raise ValueError("an expression must be of the form 'path/to/key=...'") from None

        if re.match(r"^\[.+\]$", value):
            values = [v.strip() for v in value.strip("[]").split(",")]
            substitutions += [(kpath, values)]
        else:
            substitutions += [(kpath, [value])]

    fpath = pathlib.Path(fhandle.name)
    onameprefix = fpath.stem
    onamesuffix = fpath.suffix

    # first generate a list of list of tuples [ [(key/a, 10), (key/a, 20), ...], [(key/b, 100), ...], ...]
    for substtuple in itertools.product(*[[(k, v) for v in values] for k, values in substitutions]):
        # ... then iterate over the cartesian product
        curr_tree = deepcopy(tree)  # create a full copy of the initial tree
        onameparts = []  # output name parts
        for kpath, value in substtuple:
            ref = curr_tree

            sections = kpath.split("/")
            for section in sections[:-1]:
                if isinstance(ref, list):
                    section = int(section)  # if we encounter a list, convert the respective path element
                ref = ref[section]  # exploit Python using references into dicts/lists

            attr = sections[-1]
            if isinstance(ref, (list, tuple)):
                attr = int(attr)

            if isinstance(ref, tuple):
                # we only get tuples for keywords which can take multiple words, hence this should be safe
                lref = list(ref)
                lref[attr] = value
                ref = tuple(lref)
            else:
                ref[attr] = value

            # take only the attribute name
            onameparts += [f"{attr}_{value}"]

        opath = pathlib.Path(f"{onameprefix}-{'-'.join(onameparts)}{onamesuffix}")
        print(f"Writing '{opath}'...")

        with opath.open("w") as fouthandle:
            fouthandle.write(f"! Generated with the CP2K input tool cp2kgen v{__version__}\n! ")
            fouthandle.write(" \\\n!   ".join(f"'{arg}'" for arg in sys.argv))
            fouthandle.write("\n")
            cp2k_generator = CP2KInputGenerator()
            for line in cp2k_generator.line_iter(curr_tree):
                fouthandle.write(f"{line}\n")



import logging
import sys

import click

try:
    import pygls  # noqa: F401

    from cp2k_input_tools.ls import cp2k_server
except ImportError:
    print(
        """Could not import the pygls package. You have to install the cp2k-input-tools with the 'lsp' extra:

pip install cp2k-input-tools[lsp]
    """
    )
    sys.exit(1)


@click.command()
@click.option("--tcp", is_flag=True, default=False, help="use TCP server instead of stdio")
@click.option("--host", default="127.0.0.1", type=str, help="bind to this address")
@click.option("--port", type=int, default=2087, help="bind to this port")
@click.option("--debug", is_flag=True, default=False, help="write a cp2kls.log file")
def cp2k_language_server(tcp, host, port, debug):
    """Language Server Protocol (LSP) implementation for CP2K input files"""
    if debug:
        logging.basicConfig(filename="cp2kls.log", level=logging.DEBUG, filemode="w")

    if tcp:
        cp2k_server.start_tcp(host, port)
    else:
        cp2k_server.start_io()



from typing import Union

from pygls.lsp.methods import (
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_CLOSE,
    TEXT_DOCUMENT_DID_OPEN,
)
from pygls.lsp.types import (
    Diagnostic,
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    Position,
    Range,
)
from pygls.server import LanguageServer

from .parser import CP2KInputParser
from .parser_errors import ParserError
from .tokenizer import TokenizerError


def _validate(ls, params: Union[DidChangeTextDocumentParams, DidCloseTextDocumentParams, DidOpenTextDocumentParams]):
    ls.show_message_log("Validating CP2K input...")

    diagnostics = []

    text_doc = ls.workspace.get_document(params.text_document.uri)
    parser = CP2KInputParser()

    with open(text_doc.path, "r") as fhandle:
        try:
            parser.parse(fhandle)
        except (TokenizerError, ParserError) as exc:
            ctx = exc.args[1]
            line = ctx.line.rstrip()

            msg = f"Syntax error: {exc.args[0]} ({exc.__cause__})"

            linenr = ctx.linenr - 1
            colnr = ctx.colnr

            if colnr is not None:
                count = 0  # number of underline chars after (positiv) or before (negative) the marker if ref_colnr given
                nchars = colnr  # relevant line length

                if ctx.ref_colnr is not None:
                    count = ctx.ref_colnr - ctx.colnr
                    nchars = min(ctx.ref_colnr, ctx.colnr)  # correct if ref comes before

                if ctx.colnrs:
                    # shift by the number of left-stripped ws
                    # ctx.colnrs contains the left shift for each possibly continued line
                    nchars += ctx.colnrs[0]  # assume no line-continuation for now

                # at least do one context
                count = max(1, count)

                erange = Range(
                    start=Position(line=linenr, character=colnr + 1 - count), end=Position(line=linenr, character=colnr + 1)
                )

            else:
                erange = Range(start=Position(line=linenr, character=1), end=Position(line=linenr, character=len(line)))

            diagnostics += [Diagnostic(range=erange, message=msg, source=type(ls).__name__)]

    ls.publish_diagnostics(text_doc.uri, diagnostics)


def setup_cp2k_ls_server(server):
    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    def did_change(ls, params: DidChangeTextDocumentParams):
        """Text document did change notification."""
        _validate(ls, params)

    @server.feature(TEXT_DOCUMENT_DID_CLOSE)
    def did_close(ls: LanguageServer, params: DidCloseTextDocumentParams):
        """Text document did close notification."""
        pass

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    async def did_open(ls, params: DidOpenTextDocumentParams):
        """Text document did open notification."""
        _validate(ls, params)


cp2k_server = LanguageServer()
setup_cp2k_ls_server(cp2k_server)



# Copyright (c), Tiziano Müller
# SPDX-License-Identifier: MIT

import itertools
from decimal import Decimal, InvalidOperation
from typing import Iterator, List, Sequence

from pydantic import BaseModel, Extra, Field, root_validator

from .utils import DatafileIterMixin, FromDictMixin, dformat


class PseudopotentialDataLocal(BaseModel):
    r: Decimal
    coefficients: List[Decimal] = Field(..., alias="coeffs")

    class Config:
        extra = "forbid"
        allow_population_by_field_name = True


class PseudopotentialDataNonLocal(BaseModel):
    r: Decimal
    nproj: int
    coefficients: List[Decimal] = Field(..., alias="coeffs")

    @root_validator
    def check_coefficients(cls, values):
        assert (
            len(values["coefficients"]) == values["nproj"] * (values["nproj"] + 1) // 2
        ), "invalid number of coefficients for non-local projection"
        return values

    class Config:
        extra = "forbid"
        allow_population_by_field_name = True


class PseudopotentialDataNLCC(BaseModel, extra=Extra.forbid):
    """Nonlinear Core Correction data"""

    r: Decimal
    n: int
    c: Decimal


class PseudopotentialData(BaseModel, DatafileIterMixin, FromDictMixin, extra=Extra.forbid):
    element: str
    identifiers: List[str]
    n_el: List[int]
    local: PseudopotentialDataLocal
    non_local: List[PseudopotentialDataNonLocal]
    nlcc: List[PseudopotentialDataNLCC] = Field(default_factory=list)

    @classmethod
    def from_lines(cls, lines: Sequence[str]) -> "PseudopotentialData":
        """
        Parse a single CP2K pseudopotential entry

        :param lines: List of strings where each string is a line from the original file
        :return:      A dictionary containing the element, name, tags, aliases, n_el, local, non_local attributes
        """

        # pylint: disable=too-many-locals

        # the first line contains the element and one or more identifiers/names
        identifiers = lines[0].split()
        element = identifiers.pop(0)

        # The second line contains the number of electrons for each angular momentum
        n_el = [int(n) for n in lines[1].split()]

        # The third line contains the local part in the format
        #   <radius> <nfuncs> [<func-coeff-1> [<func-coeff-2> ...]]
        r_loc_s, nexp_ppl_s, *cexp_ppl_s = lines[2].split()

        local = PseudopotentialDataLocal(r=Decimal(r_loc_s), coefficients=[Decimal(f) for f in cexp_ppl_s])

        if int(nexp_ppl_s) != len(local.coefficients):
            raise ValueError("less coefficients found than expected while parsing the block")

        prj_ppnl: List[PseudopotentialDataNonLocal] = []
        nlcc: List[PseudopotentialDataNLCC] = []
        nline = 3

        if len(lines) > nline:
            if lines[nline].startswith("NLCC"):
                try:
                    _, n_nlcc_str = lines[nline].split()
                    n_nlcc = int(n_nlcc_str)
                    nline += 1
                    while len(nlcc) < n_nlcc:
                        nlcc_r, nlcc_n, nlcc_c = lines[nline].split()
                        nlcc.append(PseudopotentialDataNLCC(r=Decimal(nlcc_r), n=int(nlcc_n), c=Decimal(nlcc_c)))
                        nline += 1
                except IndexError as exc:
                    raise ValueError("premature end-of-lines while reading the NLCC parameters") from exc
                except (ValueError, InvalidOperation) as exc:
                    raise ValueError(f"parsing values failed for line: '{lines[nline]}'") from exc

            nprj = int(lines[nline])
            nline += 1  # start processing the non-local function blocks (if any) at this line
            while len(prj_ppnl) < nprj:
                try:
                    r_nprj_s, nprj_ppnl_s, *hprj_ppnl_strings = lines[nline].split()
                    hprj_ppnl = [Decimal(c) for c in hprj_ppnl_strings]
                except IndexError as exc:
                    raise ValueError("premature end-of-lines while reading a block of non-local projectors") from exc

                nline += 1

                nprj_ppnl = int(nprj_ppnl_s)
                ncoeffs = nprj_ppnl * (nprj_ppnl + 1) // 2  # number of elements in the upper triangular matrix

                # the matrix may be distributed over multiple lines, add those values as well
                while len(hprj_ppnl) < ncoeffs:
                    try:
                        hprj_ppnl += [Decimal(c) for c in lines[nline].split()]
                    except IndexError as exc:
                        raise ValueError("premature end-of-lines while reading coefficients of non-local projects") from exc
                    except InvalidOperation as exc:
                        raise ValueError(f"parsing coefficients failed for line: '{lines[nline]}'") from exc
                    nline += 1

                if len(hprj_ppnl) > ncoeffs:
                    raise ValueError("unknown format of the non-local projector coefficients")

                prj_ppnl.append(
                    PseudopotentialDataNonLocal(
                        r=Decimal(r_nprj_s),
                        nproj=nprj_ppnl,  # store for convenience
                        coefficients=hprj_ppnl,  # upper triangular matrix
                    )
                )

        return cls(element=element, identifiers=identifiers, n_el=n_el, local=local, non_local=prj_ppnl, nlcc=nlcc)

    def cp2k_format_line_iter(self) -> Iterator[str]:
        """Generate lines of strings from this PP in the format expected by CP2K."""

        def all_radii():
            return itertools.chain((self.local.r,), (nl.r for nl in self.non_local))

        def all_coeffs():
            return itertools.chain(self.local.coefficients, (c for nl in self.non_local for c in nl.coefficients))

        i_fmt = 4

        r_max_exp = -min(r.as_tuple().exponent for r in all_radii())
        r_max_len = max(6 + r_max_exp, *(len(f"{r:.{r_max_exp}f}") for r in all_radii()))

        try:
            c_max_exp = -min(c.as_tuple().exponent for c in all_coeffs())
            c_max_len = max(6 + c_max_exp, *(len(f"{c:.{c_max_exp}f}") for c in all_coeffs()))
        except (ValueError, TypeError):  # for all-electron pseudos all_coeffs is an empty sequence and min/max start to fail
            pass

        yield f"{self.element:2} {' '.join(n for n in self.identifiers)}"
        yield " " + " ".join(f"{i:{i_fmt}}" for i in self.n_el)
        yield (
            f" {dformat(self.local.r, r_max_exp, r_max_len)} {len(self.local.coefficients):{i_fmt}} "
            + " ".join(f"{dformat(c, c_max_exp, c_max_len)}" for c in self.local.coefficients)
        ).rstrip()

        if self.nlcc:
            yield f" NLCC {len(self.nlcc):{i_fmt}}"

            r_nlcc_max_exp = -min(nlcc.r.as_tuple().exponent for nlcc in self.nlcc)
            r_nlcc_max_len = max(6 + r_nlcc_max_exp, *(len(f"{nlcc.r:.{r_nlcc_max_exp}f}") for nlcc in self.nlcc))

            c_nlcc_max_exp = -min(nlcc.c.as_tuple().exponent for nlcc in self.nlcc)
            c_nlcc_max_len = max(6 + c_nlcc_max_exp, *(len(f"{nlcc.c:.{c_nlcc_max_exp}f}") for nlcc in self.nlcc))

            for nlcc in self.nlcc:
                yield (
                    f" {dformat(nlcc.r, r_nlcc_max_exp, r_nlcc_max_len)} {nlcc.n:{i_fmt}}"
                    f" {dformat(nlcc.c, c_nlcc_max_exp, c_nlcc_max_len)}"
                ).rstrip()

        yield f" {len(self.non_local):{i_fmt}}"
        nl_matrix_indent = 1 + r_max_len + 1 + i_fmt + 1

        for nonl in self.non_local:
            # print the first N (=nproj) coefficients (first row of the matrix)
            yield (
                f" {dformat(nonl.r, r_max_exp, r_max_len)} {nonl.nproj:{i_fmt}} "
                + " ".join(f"{dformat(c, c_max_exp, c_max_len)}" for c in nonl.coefficients[: nonl.nproj])
            ).rstrip()

            # for a non-scalar non-empty matrix, print the rest of the coefficients
            for nrow in range(1, nonl.nproj):
                scol = nrow * nonl.nproj - nrow * (nrow - 1) // 2
                ecol = scol + nonl.nproj - nrow
                yield (
                    " " * (nl_matrix_indent + nrow * (c_max_len + 1))
                    + " ".join(f"{dformat(c, c_max_exp, c_max_len)}" for c in nonl.coefficients[scol:ecol])
                ).rstrip()



"""
Parsers and serializers for the Basis Set format used by CP2K
"""

import re
from decimal import Decimal
from typing import Iterator, List, Optional, Sequence, Tuple

from pydantic import BaseModel, Extra

from .utils import SYM2NUM, DatafileIterMixin, FromDictMixin, dformat

N_VAL_EL_MATCH = re.compile(r"q(?P<nvalel>\d+)$")


class BasisSetCoefficients(BaseModel, extra=Extra.forbid):
    """A 'shell' in one single basis set"""

    n: int
    l: List[Tuple[int, int]]
    coefficients: List[List[Decimal]]


class BasisSetData(BaseModel, DatafileIterMixin, FromDictMixin, extra=Extra.forbid):
    """Basis set data for a single element"""

    element: str
    identifiers: List[str]
    n_el: Optional[int]
    blocks: List[BasisSetCoefficients]

    @classmethod
    def from_lines(cls, lines: Sequence[str]) -> "BasisSetData":
        # the first line contains the element and one or more identifiers/names
        identifiers = lines[0].split()
        element = identifiers.pop(0)

        n_el: Optional[int] = None

        try:
            n_el = next(int(m["nvalel"]) for m in (N_VAL_EL_MATCH.search(n) for n in identifiers) if m)
        except StopIteration:
            pass

        # the ALL* tags indicate an all-electron basis set, but they might be ambigious,
        # ignore them if we found an explicit #(val.el.) spec already
        if not n_el and any(kw in identifiers for kw in ("ALL", "ALLELECTRON")):
            n_el = SYM2NUM[element]

        # The second line contains the number of sets, conversion to int ignores any whitespace
        n_blocks = int(lines[1])

        nline = 2

        blocks = []

        # go through all blocks containing different sets of orbitals
        for nblock in range(n_blocks):
            # get the quantum numbers for this set, formatted as follows:
            # n lmin lmax nexp nshell(lmin) nshell(lmin+1) ... nshell(lmax-1) nshell(lmax)
            # ignore everything after nshell(lmax) on the same line (as CP2K does)
            tokens = lines[nline].split()
            qn_n, qn_lmin, qn_lmax, nexp = [int(qn) for qn in tokens[:4]]
            ncoeffs = [int(qn) for qn in tokens[4 : 5 + qn_lmax - qn_lmin]]  # noqa: E203

            nline += 1

            try:
                coefficients = [[Decimal(c) for c in lines[nline + n].split()] for n in range(nexp)]
            except IndexError:
                raise ValueError(f"Not enough exponents found. Expected {nexp} lines for block {nblock+1}") from None

            blocks.append(
                BasisSetCoefficients(
                    n=qn_n,
                    l=[(lqn, nl) for lqn, nl in zip(range(qn_lmin, qn_lmax + 1), ncoeffs)],
                    coefficients=coefficients,
                )
            )

            # advance by the number of exponents
            nline += nexp

        return cls(element=element, identifiers=identifiers, n_el=n_el, blocks=blocks)

    def cp2k_format_line_iter(self) -> Iterator[str]:
        """Generate lines of strings from this Basis Set in the format expected by CP2K."""

        yield f"{self.element:2} {' '.join(n for n in self.identifiers)}"
        yield f" {len(self.blocks):2}"  # the number of sets this basis set contains

        max_exp = -min(c.as_tuple().exponent for b in self.blocks for r in b.coefficients for c in r)
        max_len = max(len(f"{c:.{max_exp}f}") for b in self.blocks for r in b.coefficients for c in r[1:])
        max_len_exp = max(9 + max_exp, *(len(str(r[0])) for b in self.blocks for r in b.coefficients))

        for block in self.blocks:
            l_str = " ".join(f"{lqn[1]:2}" for lqn in block.l)
            yield f" {block.n:2} {block.l[0][0]:2} {block.l[-1][0]:2} {len(block.coefficients):2} {l_str}"

            for row in block.coefficients:
                yield (
                    f" {dformat(row[0], max_exp, max_len_exp)} " + " ".join(f"{dformat(c, max_exp, max_len)}" for c in row[1:])
                ).rstrip()



class ParserError(Exception):
    pass


class InvalidNameError(ParserError):
    pass


class SectionMismatchError(ParserError):
    pass


class InvalidSectionError(ParserError):
    pass


class InvalidParameterError(ParserError):
    pass


class NameRepetitionError(ParserError):
    pass


class PreprocessorError(ParserError):
    pass



import click

from cp2k_input_tools.parser import CP2KInputParser, CP2KInputParserSimplified

from . import base_dir_option, canonical_option, fhandle_argument, var_values_option


@click.command()
@fhandle_argument
@click.argument("paths", metavar="<path>", type=str, nargs=-1)
@base_dir_option
@var_values_option
@canonical_option
def cp2kget(fhandle, paths, base_dir, var_values, canonical):
    """Get values by path from a CP2K input file

    Examples for paths:

        force_eval/dft/mgrid/cutoff
    """

    if canonical:
        cp2k_parser = CP2KInputParser(base_dir=base_dir, key_trafo=str.lower)
    else:
        cp2k_parser = CP2KInputParserSimplified(base_dir=base_dir, key_trafo=str.lower)

    tree = cp2k_parser.parse(fhandle, dict(var_values))

    def _(val):
        if isinstance(val, list):
            return ", ".join(str(v) for v in val)

        return val

    for path in paths:
        sections = path.split("/")
        ref = tree
        for section in sections:
            if isinstance(ref, (list, tuple)):
                section = int(section)  # if we encounter a list, convert the respective path element
            ref = ref[section]  # exploit Python using references into dicts/lists

        print(f"{path}: {_(ref)}")



import pathlib

import click

from cp2k_input_tools.basissets import BasisSetData
from cp2k_input_tools.pseudopotentials import PseudopotentialData

from . import smart_open


@click.command()
@click.argument(
    "iformat",
    metavar="<format>",
    type=click.Choice(("basis", "basisset", "basissets", "pseudo", "pseudos", "pseudopotential", "pseudopotentials", "potentials")),
)
@click.argument(
    "fpath", metavar="[<file to lint>]", type=click.Path(dir_okay=False, allow_dash=True, path_type=pathlib.Path), default="-"
)
@click.option("-i", "--inplace", is_flag=True, help="replace the original file with the linted/prettified one")
def cp2k_datafile_lint(iformat, fpath, inplace):
    """Linter/Pretty-printer for other CP2K data formats

    Supported formats are:

        basissets (aliases: basis, basisset, basissets)
        potentials (aliases: pseudo, pseudos, pseudopotential, pseudopotentials, potentials)
    """

    if iformat in ("basis", "basisset", "basissets"):
        datafile = BasisSetData
    elif iformat in ("pseudo", "pseudos", "pseudopotential", "pseudopotentials", "potentials"):
        datafile = PseudopotentialData

    swpfile = None

    if inplace:
        assert str(fpath) != "-", "Replacing file content does not work when reading from stdin"
        swpfile = fpath.parent / f".{fpath.name}.swp"

    has_preceeding_comments = False
    with smart_open(fpath, "r") as fhandle:
        with smart_open(swpfile, "x") as fouthandle:
            for entry in datafile.datafile_iter(fhandle, keep_going=False, emit_comments=True):
                if isinstance(entry, str):
                    print(entry, file=fouthandle)
                    has_preceeding_comments = True
                else:
                    if not has_preceeding_comments:  # if there is no comment before the next entry, add at least an empty one
                        print("#", file=fouthandle)

                    for line in entry.cp2k_format_line_iter():
                        print(line, file=fouthandle)

                    has_preceeding_comments = False

    if inplace:
        swpfile.replace(fpath)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cp2k_input_tools', 'cp2k_input_tools.cli']

package_data = \
{'': ['*'], 'cp2k_input_tools': ['templates/*']}

install_requires = \
['Jinja2>=2.11.3,<4.0.0',
 'click>=7.1.2,<9',
 'pint>=0.15,<0.18',
 'pydantic>=1.8.2,<2.0.0',
 'transitions>=0.7,<0.8']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<1'],
 'lsp': ['pygls>=0.10,<0.12'],
 'yaml': ['ruamel.yaml>=0.16.5,<0.18.0']}

entry_points = \
{'console_scripts': ['cp2k-datafile-lint = '
                     'cp2k_input_tools.cli.datafile_lint:cp2k_datafile_lint',
                     'cp2k-language-server = '
                     'cp2k_input_tools.cli.lsp:cp2k_language_server',
                     'cp2kgen = cp2k_input_tools.cli.cp2kgen:cp2kgen',
                     'cp2kget = cp2k_input_tools.cli.cp2kget:cp2kget',
                     'cp2klint = cp2k_input_tools.cli.lint:cp2klint',
                     'fromcp2k = cp2k_input_tools.cli.fromcp2k:fromcp2k',
                     'tocp2k = cp2k_input_tools.cli.tocp2k:tocp2k']}

setup_kwargs = {
    'name': 'cp2k-input-tools',
    'version': '0.8.1',
    'description': 'Python tools to handle CP2K input files',
    'long_description': '# cp2k-input-tools\n\n[![Build status](https://github.com/cp2k/cp2k-input-tools/actions/workflows/test.yml/badge.svg)](https://github.com/cp2k/cp2k-input-tools/actions) [![codecov](https://codecov.io/gh/cp2k/cp2k-input-tools/branch/develop/graph/badge.svg)](https://codecov.io/gh/cp2k/cp2k-input-tools) [![PyPI](https://img.shields.io/pypi/pyversions/cp2k-input-tools)](https://pypi.org/project/cp2k-input-tools/)\n\nFully validating pure-python CP2K input file parsers including preprocessing capabilities\n\nAvailable commands (also available through an API, see below):\n\n* `cp2klint` .. a CP2K input file linter\n* `fromcp2k` .. create a JSON/YAML configuration file or an [AiiDA](https://github.com/aiidateam/aiida-cp2k) run script from a CP2K input file (includes validation)\n* `tocp2k` .. convert a JSON or YAML configuration back to CP2K\'s input file format (includes validation)\n* `cp2kgen` .. generate new input files based on a given input file and expressions to change parameters programmatically\n* `cp2kget` .. get values from a CP2K input file (most likely a restart file) given a path of sections and attribute\n* `cp2k-language-server` .. a [Language Server Protocol](https://microsoft.github.io/language-server-protocol/) implementation for the CP2K input file format\n* `cp2k-datafile-lint` .. linters for other CP2K-related data file formats (like pseudos and basissets)\n\nFor a description of the JSON/YAML formats used, see below.\n\n## Requirements\n\n* Python 3.6+\n* https://pypi.org/project/transitions/\n* https://pypi.org/project/pint/\n* optional: https://pypi.org/project/ruamel.yaml/ for YAML support\n* optional: https://github.com/openlawlibrary/pygls for the cp2k-language-server\n\nFor development: https://poetry.eustace.io/ https://pytest.org/\n\n## Idea\n\n* have a pure-python CP2K input file linter with proper syntax error reporting (context, etc.)\n* a final & complete restart file parser\n* basis for an AiiDA CP2K project importer\n* testbed for alternative import formats (YAML, JSON) for CP2K\n* possible testbed for a re-implementation of the CP2K input parser itself\n\n## TODOs\n\n* parser: improve error reporting with context\n* preprocessor: don\'t lose original context when interpolating variables\n* parser: parsing the XML is slow (easily 70% of the time), pickle or generate Python code directly instead and keep XML parsing as fallback\n* parser: preserve comments when going to/from YAML\n\n# Usage\n\n## Installation\n\nYou will get most tools using simply:\n\n```console\n$ pip install cp2k-input-tools\n```\n\nFor YAML support you should use\n\n```console\n$ pip install cp2k-input-tools[yaml]\n```\n\nand for the Language Server:\n\n```console\n$ pip install cp2k-input-tools[lsp]\n```\n\n## Command Line Interface\n\nGenerate JSON, YAML or aiida-cp2k run script from a CP2K input file:\n\n```console\n$ fromcp2k --help\nusage: fromcp2k [-h] [-y] [-c] [-b BASE_DIR] [-t TRAFO] <file>\n\nConvert CP2K input to JSON (default) or YAML\n\npositional arguments:\n  <file>                CP2K input file\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -y, --yaml            output yaml instead of json\n  -c, --canonical       use the canonical output format\n  -b BASE_DIR, --base-dir BASE_DIR\n                        search path used for relative @include\'s\n  -t TRAFO, --trafo TRAFO\n                        transformation applied to key and section names (auto,\n                        upper, lower)\n```\n\nGenerate an [aiida-cp2k](https://github.com/aiidateam/aiida-cp2k) template run script:\n\n```console\n$ fromcp2k --format aiida-cp2k-calc tests/inputs/test01.inp\nAny key transformation function other than \'auto\' is ignored when generating an aiida-cp2k run script template\n\nfrom aiida.engine import run\nfrom aiida.orm import (load_code, Dict, StructureData)\n\ncp2k_code = load_code(\'...\')\n\n# Structure\nstructure = StructureData(...)\n\n# Parameters\nparameters = Dict(\n    dict={\n           "FORCE_EVAL": {\n               "DFT": {\n                   "KPOINTS": {\n                       "FULL_GRID": False,\n                       "PARALLEL_GROUP_SIZE": -1,\n                       "SCHEME": "MONKHORST-PACK 3 3 3",\n                       "SYMMETRY": False,\n                   },\n                   "MGRID": {\n                       "CUTOFF": 1000.0,\n                       "REL_CUTOFF": 100.0,\n                   },\n                   "POISSON": {\n                       "PERIODIC": "XYZ",\n                   },\n                   "PRINT": {\n                       "OVERLAP_CONDITION": {\n                           "_": "ON",\n                           "1-NORM": True,\n                           "DIAGONALIZATION": True,\n                       },\n                   },\n                   "QS": {\n                       "EPS_DEFAULT": 1e-16,\n                       "EXTRAPOLATION": "USE_GUESS",\n                       "METHOD": "GAPW",\n                   },\n                   "SCF": {\n                       "SMEAR": {\n                           "_": True,\n                           "ELECTRONIC_TEMPERATURE": 300.0,\n                           "METHOD": "FERMI_DIRAC",\n                       },\n                       "ADDED_MOS": 40,\n                       "EPS_SCF": 1e-08,\n                       "MAX_SCF": 50,\n                   },\n                   "XC": {\n                       "XC_FUNCTIONAL": {\n                           "_": "PBE",\n                       },\n                   },\n                   "BASIS_SET_FILE_NAME": "./BASIS_SETS",\n                   "POTENTIAL_FILE_NAME": "./POTENTIALS",\n               },\n               "SUBSYS": {\n                   "CELL": {\n                       "CELL_REF": {\n                           "A": "4.32947291598 0.0 0.0",\n                           "B": "2.16473645799 3.7494335304 0.0",\n                           "C": "2.16473645799 1.24981118034 3.53499983838",\n                           "PERIODIC": "XYZ",\n                       },\n                       "A": "4.07419 0.0 0.0",\n                       "B": "2.037095 3.52835204 0.0",\n                       "C": "2.037095 1.17611735 3.32656221",\n                       "PERIODIC": "XYZ",\n                   },\n                   "KIND": [\n                       {\n                       "_": "Ge",\n                       "ELEMENT": "Ge",\n                       "POTENTIAL": "ALL-q32",\n                       "BASIS_SET": "ORB pob-TZVP",\n                       },\n                   ],\n                   "TOPOLOGY": {\n                       "COORD_FILE_NAME": "./struct.xyz",\n                       "COORD_FILE_FORMAT": "XYZ",\n                   },\n               },\n               "METHOD": "QUICKSTEP",\n           },\n           "GLOBAL": {\n               "PRINT_LEVEL": "MEDIUM",\n               "PROJECT_NAME": "fatman.calc",\n               "RUN_TYPE": "ENERGY",\n           },\n    })\n\n# Construct process builder.\nbuilder = cp2k_code.get_builder()\nbuilder.structure = structure\nbuilder.parameters = parameters\nbuilder.code = cp2k_code\nbuilder.metadata.options.resources = {\n    "num_machines": 1,\n    "num_mpiprocs_per_machine": 1,\n}\nbuilder.metadata.options.max_wallclock_seconds = 1 * 3 * 60\n\nrun(builder)\n```\n\n\nGenerate a CP2K input file from a JSON or YAML:\n\n```console\n$ tocp2k --help\nusage: tocp2k [-h] [-y] <file>\n\nConvert JSON or YAML input to CP2K\n\npositional arguments:\n  <file>      JSON or YAML input file\n\noptional arguments:\n  -h, --help  show this help message and exit\n  -y, --yaml\n```\n\nLint a CP2K input file:\n\n```console\n$ cp2klint tests/inputs/unterminated_var.inp\nSyntax error: unterminated variable, in tests/inputs/unterminated_var.inp:\nline   36: @IF ${HP\n               ~~~~^\n```\n\nGenerate input files for a `CUTOFF` convergence study (multiple expressions will be combined as a cartesian product):\n\n```console\n$ cp2kgen tests/inputs/NaCl.inp "force_eval/dft/mgrid/cutoff=[800,900,1000]"\nWriting \'NaCl-cutoff_800.inp\'...\nWriting \'NaCl-cutoff_900.inp\'...\nWriting \'NaCl-cutoff_1000.inp\'...\n$ diff -Naurb NaCl-cutoff_800.inp NaCl-cutoff_900.inp\n--- NaCl-cutoff_800.inp\t2019-10-21 18:52:09.994323474 +0200\n+++ NaCl-cutoff_900.inp\t2019-10-21 18:52:10.680996641 +0200\n@@ -69,7 +69,7 @@\n       POTENTIAL_FILE_NAME ALL_POTENTIALS\n       &MGRID\n          REL_CUTOFF 80.0\n-         CUTOFF 800\n+         CUTOFF 900\n          NGRIDS 6\n       &END MGRID\n       &XC\n```\n\nGet a value from a CP2K input file, for example a `RESTART` file generated in a cell optimization:\n\n```console\n$ cp2kget tests/inputs/NaCl.inp "force_eval/subsys/cell/a/0"\nforce_eval/subsys/cell/a/0: 5.64123539364476\n```\n\n## API\n\nConvert a CP2K input file to a nested Python dictionary:\n\n```python\nfrom cp2k_input_tools.parser import CP2KInputParser, CP2KInputParserSimplified\n\ncanonical = False\n\nif canonical:\n    parser = CP2KInputParser()\nelse:\n    parser = CP2KInputParserSimplified()\n\nwith open("project.inp") as fhandle:\n    tree = parser.parse(fhandle)\n```\n\nConvert a nested Python dictionary back to a CP2K input file:\n\n```python\nfrom cp2k_input_tools.generator import CP2KInputGenerator\n\ngenerator = CP2KInputGenerator()\n\ntree = {"global": {}}  # ... the input tree\n\nwith open("project.inp", "w") as fhandle:\n    for line in generator.line_iter(tree):\n        fhandle.write(f"{line}\\n")\n```\n\n## Language Server Protocol\n\nThe executable providing the language server is: `cp2k-language-server`\n\nFor `vim` you need a plugin to be able to use language servers. One such plugin is [ALE](https://github.com/dense-analysis/ale) for which you can create in its directory the file `ale_linters/cp2k/language_server.vim` with the content\n\n```vim\ncall ale#Set(\'cp2k_lsp_executable\', \'cp2k-language-server\')\n\nfunction! ale_linters#cp2k#language_server#GetProjectRoot(buffer) abort\n    let l:git_path = ale#path#FindNearestDirectory(a:buffer, \'.git\')\n\n    return !empty(l:git_path) ? fnamemodify(l:git_path, \':h:h\') : \'\'\nendfunction\n\ncall ale#linter#Define(\'cp2k\', {\n\\   \'name\': \'language_server\',\n\\   \'lsp\': \'stdio\',\n\\   \'executable\': {b -> ale#Var(b, \'cp2k_lsp_executable\')},\n\\   \'project_root\': function(\'ale_linters#cp2k#language_server#GetProjectRoot\'),\n\\   \'command\': \'%e\',\n\\})\n```\nAfterwards you must set the filetype when editing a CP2K input file to `cp2k` to get it running. This can be done explicitly using `:set filetype=cp2k`.\n\n\n# The CP2K JSON and YAML formats\n\nA reference to the CP2K input format can be found here: https://manual.cp2k.org/\n\n## Canonical format\n\nFor everything except the pre-processor capabilities (`@IF/@ENDIF/$var/@SET`) there is a canonical one-to-one mapping of the CP2K input format to either JSON or YAML:\n\n* repeatable sections are mapped to dictionaries\n* keywords or subsections are key/value entries in sections\n* all repeatable elements (sections and keywords) are mapped to lists of their respective mapped datatype\n* section parameters are mapped to a special key named `_`\n* default section keywords are mapped to a special key name `*`\n* sections in JSON or YAML must be prefixed to avoid double definition of a key in case of same name for a section and a keyword (like the `POTENTIAL` in `KIND`), to avoid quotation marks, instead of CP2K\'s `&` we are using the `+`\n* keyword values are mapped based on their datatypes: a list of values is always mapped to a list of their respective datatypes\n\nThe following example input:\n\n```\n&GLOBAL\n   PRINT_LEVEL MEDIUM\n   PROJECT test\n   RUN_TYPE ENERGY\n&END GLOBAL\n&FORCE_EVAL\n   METHOD Quickstep\n   &DFT\n      BASIS_SET_FILE_NAME "./BASIS_SETS"\n      POTENTIAL_FILE_NAME ./POTENTIALS\n      &XC\n         &XC_FUNCTIONAL PBE\n         &END XC_FUNCTIONAL\n      &END XC\n   &END DFT\n   &SUBSYS\n      &CELL\n         A [angstrom] 4.07419 0.0 0.0\n         B [angstrom] 2.037095 3.52835204 0.0\n         C [angstrom] 2.037095 1.17611735 3.32656221\n         PERIODIC XYZ\n      &END CELL\n      &KIND Ge\n         ELEMENT Ge\n         POTENTIAL ALL-q32\n         BASIS_SET ORB pob-TZVP\n      &END KIND\n      &TOPOLOGY\n         COORD_FILE ./struct.xyz\n         COORD_FILE_FORMAT XYZ\n      &END TOPOLOGY\n   &END SUBSYS\n&END FORCE_EVAL\n```\n\nwould generate the (canonical) JSON:\n\n```json\n{\n  "+global": {\n    "print_level": "medium",\n    "project_name": "test",\n    "run_type": "energy"\n  },\n  "+force_eval": [\n    {\n      "method": "quickstep",\n      "+DFT": {\n        "basis_set_file_name": [\n          "./BASIS_SETS"\n        ],\n        "potential_file_name": "./POTENTIALS"\n      },\n      "+XC": {\n        "+xc_functional": {\n          "_": "PBE"\n        }\n      },\n      "+subsys": {\n        "cell": {\n          "A": [ 4.07419, 0, 0 ],\n          "B": [ 2.037095, 3.52835204, 0 ],\n          "C": [ 2.037095, 1.17611735, 3.32656221 ],\n          "periodic": "XYZ"\n        },\n        "+kind": [\n          {\n            "_": "Ge",\n            "element": "Ge",\n            "potential": "ALL-q32",\n            "basis_set": [\n              [ "ORB", "pob-TZVP" ]\n            ]\n          }\n          ],\n        "+topology": {\n          "coord_file_name": "./struct.xyz",\n          "coord_file_format": "XYZ"\n        }\n      }\n    }\n  ]\n}\n```\n\n*Caveats*:\n\n* the full input format needs be known and is being loaded from a bundled `cp2k_input.xml`\n* the YAML/JSON is quiet verbose and one has to know exactly which keywords can be repeated\n\nWhile there is no solution to remedy the first caveat, the second can be solved with the simplified output format\n\n## Simplified format\n\nStill based on the canonical format the simplified format relaxes some of the rules\n\n1. a section must only be prefixed with a `+` if a keyword with the same name is present at the same time in the same section (since we can figure out whether the user wanted to specify the section or the keyword by inspecting the value for the key: `dict` for a section)\n2. if a repeated keyword or section contains only one entry, the list can be omitted (in case of ambiguity priority is given to multiple values per keyword rather than keyword repetition)\n3. sections with default parameters can be formulated as dictionaries, as long as the default parameter values are unique and do not match section keyword or subsection names\n\nthe example from before in the simplified format:\n\n```json\n{\n  "global": {\n    "print_level": "medium",\n    "project_name": "test",\n    "run_type": "energy"\n  },\n  "force_eval": {\n    "method": "quickstep",\n    "DFT": {\n      "basis_set_file_name": "./BASIS_SETS",\n      "potential_file_name": "./POTENTIALS"\n    },\n    "xc": {\n      "xc_functional": {\n        "_": "PBE"\n      }\n    },\n    "subsys": {\n      "cell": {\n        "A": [ 4.07419, 0, 0 ],\n        "B": [ 2.037095, 3.52835204, 0 ],\n        "C": [ 2.037095, 1.17611735, 3.32656221 ],\n        "periodic": "XYZ"\n      },\n      "kind": {\n        "_": "Ge",\n        "element": "Ge",\n        "potential": "ALL-q32",\n        "basis_set": [ "ORB", "pob-TZVP" ]\n      },\n      "topology": {\n        "coord_file_name": "./struct.xyz",\n        "coord_file_format": "XYZ"\n      }\n    }\n  }\n}\n```\n\nor in YAML (with simplification rule #3 applied):\n\n```yaml\nglobal:\n  print_level: medium\n  project_name: test\n  run_type: energy\nforce_eval:\n  DFT:\n    basis_set_file_name: ./BASIS_SETS\n    potential_file_name: ./POTENTIALS\n  XC:\n    xc_functional:\n      _: PBE  # this can NOT be simplified since PBE could also be a subsection of xc_functional\n  method: quickstep\n  subsys:\n    cell:\n      A: [ 4.07419, 0.0, 0.0]\n      B: [ 2.037095, 3.52835204, 0.0]\n      C: [ 2.037095, 1.17611735, 3.32656221]\n      periodic: XYZ\n    kind:\n      Ge:\n        basis_set: [ORB, pob-TZVP]\n        element: Ge\n        potential: ALL-q32\n    topology:\n      coord_file_format: XYZ\n      coord_file_name: ./struct.xyz\n```\n',
    'author': 'Tiziano Müller',
    'author_email': 'tiziano.mueller@chem.uzh.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cp2k/cp2k-input-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4',
}


setup(**setup_kwargs)



import itertools
import re
import xml.etree.ElementTree as ET
from collections import Counter
from dataclasses import dataclass, field
from fractions import Fraction
from typing import Iterator, List, Optional, Tuple, Union

from . import DEFAULT_CP2K_INPUT_XML
from .keyword_helpers import UREG, Keyword
from .parser_errors import (
    InvalidNameError,
    InvalidParameterError,
    InvalidSectionError,
    NameRepetitionError,
    SectionMismatchError,
)
from .preprocessor import CP2KPreprocessor
from .tokenizer import COMMENT_CHARS, Context, TokenizerError

_SECTION_MATCH = re.compile(r"&(?P<name>[\w\-_]+)\s*(?P<param>.*)")
_KEYWORD_MATCH = re.compile(r"(?P<name>[\w\-_]+)\s*(?P<value>.*)")


@dataclass
class Section:
    name: str
    node: ET.Element
    subsections: List["Section"] = field(default_factory=list)
    keywords: List[Keyword] = field(default_factory=list)
    param: Union[int, float, str, bool, None] = None
    repeats: bool = False

    def subsections_by_name(self, name) -> Iterator["Section"]:
        yield from (s for s in self.subsections if s.name == name)

    def keywords_by_name(self, name) -> Iterator[Keyword]:
        yield from (k for k in self.keywords if k.name == name)

    @property
    def keyword_names(self) -> Iterator[Optional[str]]:
        yield from (n.text for n in self.node.iterfind("./KEYWORD/NAME"))

    @property
    def section_names(self) -> Iterator[Optional[str]]:
        yield from (n.text for n in self.node.iterfind("./SECTION/NAME"))

    def find_node_by_name(self, tag, name):
        """return the node matching the given name or None"""
        for node in self.node.iterfind(f"./{tag}"):
            # ElementTree does not have a parent relationship,
            # hence the double loop
            for sub in node.iterfind("./NAME"):
                if sub.text == name.upper():
                    return node

        return None


class CP2KInputParser:
    def __init__(self, xmlspec=DEFAULT_CP2K_INPUT_XML, base_dir=".", key_trafo=str.lower):
        """
        The CP2K input parser.

        :param xmlspec: Path to the `cp2k_input.xml` file generated with `cp2k --xml`
        :param base_dir: The base directory to be used for resolving `@include` directives
        :param key_trafo: A function object used for mangling key names, must treat input case-insensitive
        """

        # schema:
        self._spec = ET.parse(xmlspec)

        # datatree being generated:
        self._tree = None
        self._treerefs = []  # initializing to empty will make nested_dict return {} if nothing was parsed yet

        self._key_trafo = key_trafo
        self._base_inc_dir = base_dir

    def _add_tree_section(self, section_name, repeats, node):
        if not repeats and any(s.name == section_name for s in self._treerefs[-1].subsections):
            # TODO: the user possibly specified an alias, but here we only return the matching key
            raise InvalidNameError(f"the section '{section_name}' can not be defined multiple times", Context())

        self._treerefs[-1].subsections += [Section(section_name, repeats=repeats, node=node)]
        self._treerefs += [self._treerefs[-1].subsections[-1]]

    def _parse_as_section(self, line):
        match = _SECTION_MATCH.match(line)

        section_name = match.group("name").upper()
        section_param = match.group("param")

        if section_name == "END":
            section_param = section_param.rstrip()

            if section_param and section_param.upper() not in [e.text for e in self._treerefs[-1].node.iterfind("./NAME")]:
                raise SectionMismatchError("could not match open section with name: {section_param}", Context())

            # if the END param was a match or none was specified, go a level up
            self._treerefs.pop()
            return

        # check all section nodes for matching names or aliases
        section_node = self._treerefs[-1].find_node_by_name("SECTION", section_name)
        if section_node is None:
            raise InvalidSectionError(f"invalid section '{section_name}'", Context())

        repeats = True if section_node.get("repeats") == "yes" else False

        self._add_tree_section(section_name, repeats, section_node)

        # check whether we got a parameter for the section and validate it
        if section_param and not section_param.startswith(COMMENT_CHARS):
            param_node = section_node.find("./SECTION_PARAMETERS")
            if param_node:  # validate the section parameter like a kw datatype
                # there is no way we get a second section parameter, assign directly
                self._treerefs[-1].param = Keyword.from_string(param_node, section_param).values
            else:
                raise InvalidParameterError(
                    f"section parameters given for non-parametrized section '{section_name}': {section_param}", Context()
                )

    def _add_tree_keyword(self, kw):
        if not kw.repeats and any(k.name == kw.name for k in self._treerefs[-1].keywords):
            # TODO: the user possibly specified an alias, but here we only return the matching key
            raise NameRepetitionError(f"the keyword '{kw.name}' can only be mentioned once")

        self._treerefs[-1].keywords += [kw]

    def _parse_as_keyword(self, line):
        match = _KEYWORD_MATCH.match(line)

        kw_name = match.group("name").upper()
        kw_value = match.group("value")

        kw_node = self._treerefs[-1].find_node_by_name("KEYWORD", kw_name)

        # if no keyword with the given name has been found, check for a default keyword for this section
        if kw_node is None:
            kw_node = self._treerefs[-1].find_node_by_name("DEFAULT_KEYWORD", "DEFAULT_KEYWORD")
            if kw_node is not None:  # for default keywords, the whole line is the value
                kw_value = line

        if kw_node is None:
            raise InvalidNameError(f"invalid keyword '{kw_name}' specified and no default keyword for this section", Context())

        try:
            kw = Keyword.from_string(kw_node, kw_value, self._key_trafo)  # the key_trafo is needed to mangle keywords
        except InvalidParameterError as exc:
            raise InvalidParameterError(f"invalid values for keyword: {match.group('name')}", Context()) from exc

        self._add_tree_keyword(kw)

    @property
    def nested_dict(self):
        stack = self._treerefs.copy()
        tree = {}
        treerefs = [tree]

        while stack:
            currsec = stack.pop(-1)
            treeref = treerefs.pop(-1)

            for section in currsec.subsections:
                section_name = f"+{self._key_trafo(section.name)}"

                if section.repeats:
                    try:
                        treeref[section_name] += [{}]
                    except KeyError:
                        treeref[section_name] = [{}]

                    treerefs += [treeref[section_name][-1]]

                else:
                    treeref[section_name] = {}
                    treerefs += [treeref[section_name]]

                stack += [section]

            for keyword in currsec.keywords:
                keyword_name = self._key_trafo(keyword.name)

                if keyword.repeats:
                    try:
                        treeref[keyword_name] += [keyword.values]
                    except KeyError:
                        treeref[keyword_name] = [keyword.values]
                else:
                    treeref[keyword_name] = keyword.values

            if currsec.param is not None:
                treeref["_"] = currsec.param

        return tree

    def parse(self, fhandle, initial_variable_values=None):
        """Parse a CP2K input file
        :param fhandle: An open file handle. Included files will be opened/closed transparently.
        :param initial_variable_values: optional dictionary with preprocessor variable names and their initial values
        :return: A nested dictionary, the parsed option "tree"
        """

        preprocessor = CP2KPreprocessor(fhandle, self._base_inc_dir, initial_variable_values)
        self._tree = Section("/", node=self._spec.getroot())
        self._treerefs = [self._tree]

        for line in preprocessor:
            try:
                if line.startswith("&"):
                    self._parse_as_section(line)
                    continue

                self._parse_as_keyword(line)

            except (TokenizerError, InvalidParameterError, InvalidSectionError, InvalidNameError) as exc:
                exc.args[1].filename = preprocessor.fname
                exc.args[1].linenr = preprocessor.line_range[1]
                exc.args[1].colnrs = preprocessor.colnrs
                exc.args[1].line = line
                exc.args[1].section = self._treerefs[-1]
                raise

        if len(self._treerefs) > 1:
            raise SectionMismatchError(
                f"section '{self._treerefs[-1].name}' not closed",
                Context(line=line, section=self._treerefs[-1]),  # preprocessor is not valid anymore at this point
            )

        # returning the nested dictionary representation for convenience
        return self.nested_dict

    def coords(self, force_eval=0) -> Iterator[Tuple[str, Tuple[float, ...], Optional[str]]]:
        """
        Return an iterator to coordinates given in a FORCE_EVAL/SUBSYS/COORD section
        where the coordinates are proper float values and converted to Angstrom if specified in
        a different unit
        """

        try:
            coord = next(
                next(
                    next(
                        itertools.islice(self._tree.subsections_by_name("FORCE_EVAL"), force_eval, force_eval + 1)
                    ).subsections_by_name("SUBSYS")
                ).subsections_by_name("COORD")
            )
        except StopIteration:
            return

        scaled = next(coord.keywords_by_name("SCALED"), False)
        current_unit = UREG.parse_expression(next(coord.keywords_by_name("UNIT"), "ANGSTROM"), case_sensitive=False)

        for coordline in coord.keywords_by_name("*"):
            # coordinates are a series of strings according to the CP2K schema
            fields = coordline.values.split()

            name = fields[0]
            position = (float(Fraction(f)) for f in fields[1:4])  # positions can be fractions
            molname = fields[4] if len(fields) > 4 else None

            if not scaled and current_unit != UREG.angstrom:
                position = ((p * current_unit).to(UREG.angstrom).magnitude for p in position)

            yield (name, tuple(position), molname)


class CP2KInputParserSimplified(CP2KInputParser):
    """Implement structured output simplification."""

    def __init__(
        self,
        multi_value_unpack=True,
        repeated_section_unpack=True,
        level_reduction_blacklist=None,
        default_keyword_symbol="*",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._multi_value_unpack = multi_value_unpack
        self._repeated_section_unpack = repeated_section_unpack
        self._no_lvl_reduction = level_reduction_blacklist if (level_reduction_blacklist is not None) else []
        self._default_keyword_symbol = default_keyword_symbol

    def _get_value(self, keyword):
        """Conditionally unpack values"""
        if isinstance(keyword.values, (list, tuple)) and not self._multi_value_unpack:
            # this is the mode as currently employed by the aiida-cp2k plugin:
            # keywords with multiple arguments are treated as simple strings
            return " ".join(str(v) for v in keyword.values)

        return keyword.values

    @property
    def nested_dict(self):
        stack = self._treerefs.copy()
        tree = {}
        treerefs = [tree]

        while stack:
            currsec = stack.pop(-1)
            treeref = treerefs.pop(-1)

            for section in currsec.subsections:
                section_name = self._key_trafo(section.name)

                # if the section can be repeated and has a string parameter, we can possible simplify the structure
                if section.repeats:
                    # if the section is not already there, check whether to add as list or as dict with the param as subkey
                    if section_name not in treeref:
                        param_counts = Counter(s.param for s in currsec.subsections if s.name == section.name)
                        valid_keys = [n for n in itertools.chain(section.keyword_names, section.section_names)]
                        # check that the parameters are unique, strings and do not match any keywords or sections valid in that section
                        if self._repeated_section_unpack and all(
                            c == 1 and isinstance(p, str) and p.upper() not in valid_keys for p, c in param_counts.items()
                        ):
                            treeref[section_name] = {}
                        else:
                            treeref[section_name] = []

                    if isinstance(treeref[section_name], dict):
                        # if the already present section type is a section, we're using section params as keys
                        treeref[section_name][section.param] = {}
                        treerefs += [treeref[section_name][section.param]]
                    elif not any(s.name == section.name for s in currsec.subsections if s is not section) and (
                        section.name not in self._no_lvl_reduction
                    ):
                        # if the section would become a list of sections, but this is the only section with that name in
                        # the current level of the parsed tree, remove one level of the list as well
                        treeref[section_name] = {"_": section.param} if section.param is not None else {}
                        treerefs += [treeref[section_name]]
                    else:
                        treeref[section_name] += [{"_": section.param}] if section.param is not None else [{}]
                        treerefs += [treeref[section_name][-1]]

                else:
                    treeref[section_name] = {"_": section.param} if section.param is not None else {}
                    treerefs += [treeref[section_name]]

                stack += [section]

            for keyword in currsec.keywords:
                keyword_name = self._key_trafo(keyword.name)

                if keyword_name == "*":
                    keyword_name = self._default_keyword_symbol

                # if the keyword already exists as a section:
                if (keyword_name in treeref) and (
                    isinstance(treeref[keyword_name], dict)
                    or (isinstance(treeref[keyword_name], list) and isinstance(treeref[keyword_name][0], dict))
                ):
                    # prefix that sections key with a '+'
                    treeref[f"+{keyword_name}"] = treeref.pop(keyword_name)

                if keyword_name in treeref:
                    # NOTE: we don't have to check for mistakenly repeated keywords, that was already done while parsing
                    #       we are therefore not risking to append to a keyword with multiple values
                    if not isinstance(treeref[keyword_name], list):
                        # if the value is not yet a list, make it one
                        treeref[keyword_name] = [treeref[keyword_name]]

                    treeref[keyword_name] += [self._get_value(keyword)]
                else:
                    treeref[keyword_name] = self._get_value(keyword)

        return tree


class CP2KInputParserAiiDA(CP2KInputParserSimplified):
    """Implement structured output simplification as expected by aiida-cp2k as input parameter"""

    def __init__(self, *args, **kwargs):
        # aiida-cp2k uses a limited dict-based representation of the CP2K input,
        # and the simplified parser needs to be tweaked:
        # * avoid that something like "BASIS_SET ORB DZVP-MOLOPT-GTH" is unpacked
        #   into {"BASIS_SET": ("ORB", "DZVP-MOLOPT-GTH")}
        # * prevents the unpacking of repeated sections (removing the list)
        # * prevents that "KIND H" is turned into {"KIND": {"H": {"BASIS_SET": ...}}}
        #   but kept as {"KIND": {"_": "H", "BASIS_SET": ...}}} and with the option above
        #   makes it compatible with aiida-cp2k's way of CP2K input representation
        # NOTE: some CP2K input files can not be represented in this form

        super().__init__(
            key_trafo=str.upper,
            multi_value_unpack=False,
            repeated_section_unpack=False,
            level_reduction_blacklist=["KIND"],
            default_keyword_symbol=" ",
            *args,
            **kwargs,
        )



import re
import sys

import click

from cp2k_input_tools.parser import CP2KInputParser
from cp2k_input_tools.parser_errors import ParserError
from cp2k_input_tools.tokenizer import TokenizerError

from . import base_dir_option, fhandle_argument, var_values_option


@click.command()
@fhandle_argument
@var_values_option
@base_dir_option
def cp2klint(fhandle, var_values, base_dir):
    """Check the passed CP2K file for syntax errors"""

    cp2k_parser = CP2KInputParser(base_dir=base_dir)

    try:
        cp2k_parser.parse(fhandle, dict(var_values))
    except (TokenizerError, ParserError) as exc:
        ctx = exc.args[1]
        line = ctx.line.rstrip()

        print(f"Syntax error: {exc.args[0]}, in {ctx.filename}:")

        if exc.__cause__:
            print(f"              {exc.__cause__}")

        print(f"line {ctx.linenr:>4}: {line}")

        if ctx.colnr is not None:
            count = 0  # number of underline chars after (positiv) or before (negative) the marker if ref_colnr given
            nchars = ctx.colnr  # relevant line length

            if ctx.ref_colnr is not None:
                count = ctx.ref_colnr - ctx.colnr
                nchars = min(ctx.ref_colnr, ctx.colnr)  # correct if ref comes before

            if ctx.colnrs:
                # shift by the number of left-stripped ws
                # ctx.colnrs contains the left shift for each possibly continued line
                nchars += ctx.colnrs[0]  # assume no line-continuation for now

            # replace all non-ws chars with spaces:
            # - assuming a monospace font
            # - preserving other whitespace we don't know the width
            underline = re.sub(r"\S", " ", ctx.line[:nchars])

            if count >= 0:
                print(f"{str():>9}  {underline}^{str():~>{count}}")
            else:
                print(f"{str():>9}  {underline}{str():~>{-count}}^")

        if ctx.ref_line is not None:
            print("previous definition:")
            print(f"line {str():>4}: {ctx.ref_line.rstrip()}")

        sys.exit(1)

    print("All done! Happy calculating!")
